import graph,stats,math
from itertools import chain
from BDE import BDE,BDE_cont
from MDL import MDL,MDL_cont

def param_simple(l):
    mu=stats.mean(l)
    sigma=stats.stdev(l)/2
    mu0=mu-sigma
    mu1=mu+sigma
    return mu0, mu1, sigma

def param_2means(l):
    def sumvar(l,m):
        return stats.sum(map(lambda x:(x-m)**2,l))
    med=stats.lmedianscore(l)
    oldmed=med+1
    while med!=oldmed:
        l0=filter(lambda x: x<med,l)
        l1=filter(lambda x: x>med,l)
        lm=filter(lambda x: x==med,l)
        mu0=stats.mean(l0*2+lm)
        mu1=stats.mean(l1*2+lm)
        (oldmed,med)=(med,(mu0+mu1)/2)
    sigma=max(0.1**15,\
             math.sqrt((sumvar(l0+lm,mu0)+sumvar(l1,mu1)+(mu1-med)**2)/(len(l))))
    return mu0, mu1, sigma

def estprob(l,t=1.0,method=param_2means):    
    mu0,mu1,sigma = method(l)
    trans=lambda x: 1.0/(1.0+math.exp((mu1**2-mu0**2+2*(mu0-mu1)*x)/(2*t*sigma**2)))
    return map(trans,l)

epp2m= lambda l : estprob(l,method=param_2means)
epps = lambda l : estprob(l,method=param_simple)


class experiment:
    def __init__(self,name,data,perturbed,points=[],static=False):
        self._data=data
        self.perturbed=perturbed
        self.data=[]
        self.name=name
        self.points=points
        if static:
            for i,d in enumerate(self._data):
                self.data.append(self._data[i]+d)
        else:
            for i,d in enumerate(self._data[1:]):
                self.data.append(self._data[i]+d)
##         print self._data
##         print self.data

    def diffs(self):
        n = len(self.data[0])/2
        res=[]
        for p in self.data:
            s = 0
            for i in range(n):
                s+=abs(p[i]-p[i+n])
            res.append(s)
        return res

    def split_by_diff(self,diff=1):
        res = []
        for d in zip(self.data,self.diffs()):
            dat,dif=d
            if dif<=diff:
                res.append(dat)
        self.data=res
        return res
class gene:
    def __init__(self,name,n_disc,index,values=[],parents=[]):
        self.name=name
        self.n_disc=n_disc
        self.parents=parents
        self.values=values
        self.index=index
        
    def __str__(self):
        return self.name

##     def __hash__(self):
##         return self.index
         
##     def __eq__(self,other):
##         if other.__class__!=self.__class__:
##             return False
##         else:
##             return other.name==self.name
    
##     def __cmp__(self,other):
##         if other.__class__!=self.__class__:
##             return False
##         else:
##             return cmp(self.n_disc,other.n_disc)

    def __index__(self):
        return self.index

    def __add__(self,other):
        return other+self.index
        
class dataset:
    def __init__(self):
        self.n_series=0
        self.n_gen=0
        self.points=[]
        self.vertices= []
        self.n_disc =-1
        self.vertices=[]
        
    def fromFile(self,f,float_transform=epp2m):
        #number of genes in the first line
        self.n_gen=int(f.readline())
        # experiments description
        ln = f.readline().strip().split()
        self.n_series = int(ln[0])
        series = []
        #print ln
        for txt in ln[1:]:
            #print txt
            if txt.find("[")==-1:
                series.append((int(txt),[]))
            else: #perturbed genes
                #print series
                series.append((int(txt[:txt.find("[")]),
                               [int(txt[1+txt.find("["):txt.find("]")])]))
        lines = []
        self.vertices=[]
        for i in range(self.n_gen):
            ln = f.readline().strip().split()
            self.n_disc = max(0,int(ln[1]))
            self.vertices.append(gene(ln[0],self.n_disc,i))
            if self.n_disc==0:
                if float_transform: #use float_transformn if specified
                    lines.append(float_transform(map(float,ln[2:])))
                else:
                    lines.append(map(float,ln[2:]))
            else:
                lines.append(map(int,ln[2:]))
        self.points=[]
        i=0
        for s,p in series:
            d = []
            for k in range(s): 
                d.append(map(lambda x:x[i],lines))
                i+=1
            name="serie%d"%len(self.points)
            print range(len(d))
            self.points.append(experiment(name,d,p,range(len(d))))

        for v in self.vertices:
            v.parents=self.vertices
        for e in self.points:
            e.perturbed=map(lambda i: self.vertices[i],e.perturbed)
        return self

    def toNewFile(self,file):
        #perturbations
        for s in self.points:
            for g in s.perturbed:
                file.write("#perturbed\t%s\t%s\n"%(s.name,self.vertices[g]))
        #condition names\
        file.write("genes\conditions")
        for s in self.points:
            for i,d in enumerate(s._data):
                file.write("\t%s:%d"%(s.name,i))
        file.write("\n")
        #expr_data
        for i,v in enumerate(self.vertices):
            file.write(str(v))
            for s in self.points:
                for j,d in enumerate(s._data):
                    if v.n_disc>0: #discrete
                        format="\t%d"
                    else: #continuous
                        format="\t%f"
                    file.write(format%d[i])
            file.write("\n")

    def fromNewFile(self,file,float_transform=epp2m):
        pert=[]
        regul=[]
        parents={}
        disc={}
        contin=[]
        self.default=None
        ln=file.readline()
        #################
        #THE PREAMBLE
        ################
        while ln[0]=="#": #preamble
            rec=ln.strip().split()
            if rec[0]=="#perturbed":
                pert.append((rec[1],rec[2:]))
            elif rec[0]=="#regulators":
                reg=filter(lambda x:True not in map(lambda r:x in r,regul),rec[1:])
                for i in range(reg):
                    if i!=index(reg[i]):
                        reg[i]=None
                reg=filter(None,reg)
                regul.append(reg)
            elif rec[0]=="#parents":
                reg=rec[2:]
                for i in range(reg):
                    if i!=index(reg[i]):
                        reg[i]=None
                reg=filter(None,reg)
                parents[rec[1]]=reg
            elif rec[0]=="#default":
                if len(rec)==2 and rec[1]=="FLOAT":
                    self.default=[]
                else:
                    self.default=rec[1:]
            elif rec[0]=="#discrete":
                disc[rec[1]]=rec[2:]
            elif rec[0]=="#continuous":
                contin.extend[rec[1:]]
            else:
                pass #ignoring other stuff
            ln=file.readline()
        self.regulators=regul
        
        ################
        #condition names
        ################
        rec=ln.strip().split()
        #dynamic or static data?
        if len(rec[1].split(":"))==1:
            static=True
        else:
            static=False
        if static:
            conds=rec[1:]
            names=conds
        else:
            serie=rec[1].split(":")[0]
            cond=[rec[1].split(":")[1]]
            conds=[cond]
            names=[serie]
            for c in rec[2:]:
                r=c.split(":")
                if r[0]==serie:
                    cond.append(r[1])
                else:
                    serie=r[0]
                    names.append(serie)
                    cond=[r[1]]
                    conds.append(cond)
        ln=file.readline()
        ###################
        # EXPRESSION DATA #
        ###################
        self.vertices=[]
        self.vertice_names=[]
        #prepare the lists for expression data
        n_cond=sum(map(len,conds))
        exp_data=map(lambda l: [ [] for el in l],conds)
        geneindex=0
        while ln: #expression data
            rec=ln.strip().split()
            name=rec[0]
            vals=[]
            #are the values discrete or not?
            if name in contin:#did we say it's continuous?
                discrete=False
            elif name in disc.keys(): #did we say is's discrete?
                discrete=True
                vals=disc[name]
            elif self.default!=None:# did we specify a default
                if self.default==[]:
                    discrete=False
                else:
                    discrete=True
                    vals=self.default
            else: #We did not specify the type of values for this gene, try to guess it
                try:
                    i = int(rec[1])
                except ValueError: #it's not an integer
                    try:
                        i=float(rec[1])
                    except ValueError: #it's not a float -> it's an alphanumerical value
                        discrete=True
                    else:
                        discrete=False # it's a float
                else:
                    discrete=True #it's an int, let's map the strings to ints
            if discrete:
                line=rec[1:]
                if vals==[]:#we don't know what are the possible values
                    for v in line:
                        if v not in vals:
                            vals.append(v)
                    #let's sort the values
                    vals.sort()
            else: #not discrete
                line=map(float,rec[1:])
                if float_transform: #use float_transformn if specified
                    line=float_transform(line)
            
            self.vertices.append(gene(name,len(vals),geneindex,values=vals))
            self.vertice_names.append(name)
            #append this line to the expression data
            for el,l in zip(line,chain(*exp_data)):
                if discrete:
                    l.append(vals.index(el))
                else:
                    l.append(el)
            ln=file.readline()
            geneindex+=1
        ################
        #POSTPROCESSING#
        ################
        
        #SELECTING POTENTIAL PARENTS
        for v in self.vertices:
            if parents.has_key(v.name):
                v.parents=map(lambda n: self.vertices[self.vertice_names.index(n)],parents[v.name])
            else:
                #maybe there are regulators specified
                if regul==[]:
                    v.parents=self.vertices
                else:#there is a list of regulators
                    v.parents=[]
                    for r in regul:
                        if name in r:
                            break
                        else:
                            v.parents.extend(map(lambda n: self.vertices[self.vertice_names.index(n)],r))
            v.parents.sort(None,lambda g:g.n_disc)
        #process the lists of perturbed genes and make the list of experiments
        for i,(c,n) in enumerate(zip(conds,names)):
            pg = []
            for x,p in pert:
                if x==n:
                    pg.extend(map(lambda n: self.vertices[self.vertice_names.index(n)],p))# map gene names to indexes
            self.points.append(experiment(n,exp_data[i],pg,c))

        #is there a common-ground discretization?
        self.values=self.vertices[0].values
        common=True
        for g in self.vertices[1:]:
            if g.values!=self.values:
                common=False
                break
        if common:
            self.n_disc=self.vertices[0].n_disc
        else:
            self.n_disc=None
            self.values=None
        #set the number of genes
        self.n_gen=geneindex
        return self

    def sel_data(self,vert_num):
        data = []
        n_points = 0
        for e in self.points:
            if vert_num not in e.perturbed:
                data +=e.data
                n_points+=len(e.data)
        return n_points,data

    def orientation(self,par,child):
        nd,d = self.sel_data(child)
        pl = []
        cl = []
        n = len(d[0])/2
        for i in range(nd):
            pl.append(d[i][par])
            cl.append(d[i][child+n])
        sign = stats.pearsonr(pl,cl)[0]
        if sign >=0:
            return "+"
        else:
            return "-"

    
    def learn(self,score,data_factor,distrib=False):
#        if self.n_disc==0:
#            scr=eval(score+"_cont")(n_disc=self.n_disc)
#        else:
        scr=eval(score)(n_disc=self.n_disc)
        scr.data_factor=data_factor
        
        if distrib:
            from dispatch import CacheDispatch
            dispatch = CacheDispatch(scr)
            verts ={}
            from Queue import Empty
        score = 0.0
        pars ={}
        
        for x in self.vertices:
            n_points,data = self.sel_data(x)
            if distrib: #round robin dispatching
                #print i,
                dispatch.put_job(x,"learn_1",x,x.parents,data, n_points)
            else:
                par,sc = scr.learn_1(x,x.parents,data, n_points)
                #print data
                score+=sc
                #print x, map(lambda x:self.vertices[x],par)
                pars[x]=par
                
        if distrib: #clean up
            dispatch.clean()
            try:
                while True:
                    k,v = dispatch.get_result()
                    par,sc = v
                    pars[k]=par
                    score+=sc
                    #print k, map(lambda x:self.vertices[x],par)
            except Empty:
                pass

        #print score
        
        #move parents from a dictionary to a list
        par_list = []
        for v in self.vertices:
            par_list.append(pars[v])
        pars=par_list

        #print "----"
        #for par in pars:
        #    for p in par:
        #        print p,
        #    print
        #print "----"

        g = graph.graph()
        g.fromParents(self.vertices ,pars)
        for v,par in zip(self.vertices,pars):
            g.vertice_labelling[v]=v.name
            for p in par:
                g.edge_labelling[p,v]=self.orientation(p,v)
        return score,g

    def split_by_diff(self,diff=1):
        for p in self.points:
            p.split_by_diff(diff)
            
def rand_exp(n_genes,n_exp,n_disc):
    import random
    res = []
    for i in range(n_exp):
        exp = []
        for j in range(n_genes*2):
            exp.append(random.randint(0,n_disc-1))
        res.append(exp)
    return res
