#from logilab.hmm.hmm import HMM
import math
from graph import graph

class state_space:
    def __init__(self,no,sigmas):
        #are sigmas 'a list list or 'a list
        if type(sigmas[0])==type([1]):
            pass #it's 'a list list
        else:
             sigmas=[sigmas]*no#it's 'a list, let's multiplicate it 
        self.lst = [""]
        self.ng=no
        self.sigmas=sigmas
        for i in range(no):
            r2 = []
            for r in self.lst:
                for s in sigmas[i]:
                    r2.append(r+s)
            self.lst=r2

    
    def neighbours(self,st,towards=None,direct=True,genes=None):
        if genes==None:
            genes=range(self.ng)
            
        for g in genes:
            cp = st[:] # copy state
            if not direct:
                ind=range(1,len(self.sigmas[g]))
            else:
                ind=[1]
            if towards:
                if st[g]==towards[g]: #no change
                    continue
                elif self.sigmas[g].index(st[g])<self.sigmas[g].index(towards[g]): # going up
                    pass
                else:
                    ind=map(lambda x:-x,ind) #going down
            else: #going both up and down
                ind+=map(lambda x:-x,ind) 
            for i in ind:
                k = self.sigmas[g].index(st[g])+i
                if k in range(len(self.sigmas[g])):
                    yield st[:g]+self.sigmas[g][k]+st[g+1:]

    def fix_pair(self,pair,unique=True):
        s1,s2=pair
        res = []
        for s in self.neighbours(s1,towards=s2):
            res.append([s1,s])
            res.extend(self.fix_pair((s,s2)))

        if unique:
            res2=[]
            for r in res:
                if r not in res2:
                    res2.append(r)
            if res==res2:
                #print "rowne"#,res,res2
                pass
            return res2
        else:
            return res

    def direction(self,s1,s2):
        "returns the first gene that s1 and s2 disagree on and the direction (+1,-1) from s1 to s2"
        for g in range(self.ng):
            if s1[g]!=s2[g]:
                if s1[g]<s2[g]:
                    return g,1
                else:
                    return g,-1

class SLN:
    def __init__(self,no_genes,states=["0","1"]):
        self.no_genes=no_genes
        self.states=states
        self.st_spc = state_space(no_genes,states)
    
    def neighbours(self,s,i):
        return self.st_spc.neighbours(s,genes=[i],direct=False)


    def learn_dataset(self,DS,perturbed=False):
        if self.no_genes!=DS.n_gen:
            raise ValueError("wrong number of genes")
##         if len(self.states)!=DS.n_disc:
##             raise ValueError("wrong number of discretization levels")
        #get the observations out of the file
        if perturbed:
            result= graph()
            for v in range(self.no_genes):
                result.add_vert(v,DS.vertices[v])
                result.edges[v]=[]
        #print result,self.no_genes
        for v in range(self.no_genes):
            data = DS.sel_data(v)
            obs = map(lambda x: "".join(map(str,x)),data[1])
            self.learn_observations(obs)
            g = self.to_graph()
            for i,n in  enumerate(map(lambda x: x.name,DS.vertices)):
                g.vertice_labelling[i]=n
            #print v,g.parents(v)
            if not perturbed:
                return g
            else:
                for p in g.parents(v):
                    result.add_edge(p,v,g.edge_labelling[(p,v)])
        return result
                
    def pprint(self):
        print self.to_graph()
        

class SLN_fast(SLN):
    def __init__(self,*args,**kw):
        return SLN.__init__(self,*args,**kw)            
        
    def learn_observations(self,obs):
        self.obs_states=[]
        self.obs_pairs=[]
        self.freqs={}
        self.cnts={}
        os = {}
        op={}
        for i,sp in enumerate(obs):
            #print "learn",i
            s0=sp[:self.no_genes]
            s1=sp[self.no_genes:]
            for s0,s1 in self.st_spc.fix_pair((s0,s1)):
            #print sp,s0,s1
                if not os.has_key(s0):
                    os[s0]=1
                    os[s1]=1
                if not op.has_key((s0,s1)):
                    op[(s0,s1)]=1
                   
            self.obs_pairs=op.keys()
            self.obs_states=os.keys()
                
        for s in self.obs_states:
            self.cnts[s]=0
            d={}
            self.freqs[s]=d
            for s1,s2 in self.obs_pairs:
                if s1==s:
                    self.cnts[s1]+=1
                    try:
                        d[s2]+=1
                    except KeyError:
                        d[s2]=1
            
        #print self.freqs,self.obs_states,self.obs_pairs

    def freq(self,st,var):
        cnt= 0
        updown = {1:0,-1:0}
        for n in self.freqs[st]:
            #print st,n, var
            v,d = self.st_spc.direction(st,n)
            if v==var:
                updown[d]+=1
                cnt+=1

        #print st,var,self.freqs[st],cnt,updown, self.states
        
        if cnt==0:
            if st[var]==self.states[0]: # assume that from 0 we'd go down if we could
                return 0.0
            elif st[var]==self.states[-1]:
                return 1.0
            else:
                return None #raise "Houston we've got a problem"
        else:
            #print updown[1]*1.0/cnt
            return updown[1]*1.0/cnt
        
    def witness(self,s1,s2,var,thresh=0.99):
        
        v1 =self.freq(s1,var)
        v2=self.freq(s2,var)
        #print "testing",var,s1,v1,s2,v2
        if v1!=None and v2!=None:
            var2=self.st_spc.direction(s1,s2)[0]
            return  abs(v1-v2)>thresh,((v2==0.0)^(s1[var2]<s2[var2]))
        else:
            return False,False
        
    
    def to_graph(self):
        g = graph()
        for i in range(self.no_genes):
            g.add_vert(i,"ABCDEFGHJKLMNOPQRSTUVWXYZ"[i])
            g.edges[i]=[]

        for u in range(self.no_genes):
            for v in range(self.no_genes):
                found=False
                for st in self.obs_states:
                    for n in self.neighbours(st,u):
                        if n in self.obs_states:
                            w,d =self.witness(st,n,v)
                            if w:
                                #print st,n,"kraw",u,"->",v,d
                                #print self.freqs[st],self.freqs[n]
                                if d:
                                    g.add_edge(u,v,"+")
                                else:
                                    g.add_edge(u,v,"-")
                                    
                                             
                                found=True
                                break
                    if found: break
        return g
        

class SLN_hmm(SLN):
    def __init__(self,no_genes,states=["0","1"]):
        import logilab.hmm.hmm as hmm
        hmm.use__hmm=0
        SLN.__init__(self,no_genes,states)
        self.hmm = hmm.HMM(self.st_spc.lst,self.st_spc.lst)

    def find_states(self):
        res = {}
        bad_o = []
        for o in range(len(self.hmm.O_index)):
            #find a state best matching the observation index o
            maxv = 0
            best_s=-1
            for s,v in enumerate(self.hmm.B[o]):
                #print "B",o,s,v
                if v>maxv:
                    best_s=s
                    maxv=v
            if best_s in res.keys():
                bad_o.append(o)
            else:
                res[best_s]=o

        #double check

        for o in bad_o:
            print "bad",o
        return res

    def multiple_learn(self,*args,**kwds):
        #reset the probabilities
        self.hmm.setRandomProba()
        #set observation matrix to identity
        self.hmm.resetObservationProba()
        for i in range(len(self.states)**self.no_genes):
            self.hmm.B[i][i]=1.0
        return self.hmm.multiple_learn(*args,**kwds)

    def pprint(self):
        return self.hmmprint(format="%2.4f")
            
    def hmmprint(self,format="%2.4f"):
        sts =  self.find_states()
        print "   ",
        for j in range(self.hmm.N):
            try:
                print " %s  "%(self.st_spc.lst[sts[j]]),
            except:
                print "XXX",
        print
        for i in range(self.hmm.N):
            try:
                print self.st_spc.lst[sts[i]],
            except:
                print "XXX",

            for j in range(self.hmm.M):
                print format%self.hmm.A[i][j],
            print
        return self.to_graph()
    
    def images(self,st,j,thresh=0.01):
        sts =  self.find_states() # find state-observation correspondence
        obs = dict(map(lambda (x,y):(y,x),sts.items())) #and reverse it
        #print sts
        #print obs
        result = []
        i = self.st_spc.lst.index(st)
        #print i
        i = obs[i]
        #print st,i
        for k in range(len(self.st_spc.lst)): # for all possible states
            if self.hmm.A[i][k]>thresh:
                #print i,k,st,sts[k]
                r =self.states.index(self.st_spc.lst[sts[k]][j])
                if r not in result:
                    result.append(r)
        r = self.states.index(st[j])
        #print r,result
        if len(result)>1:
            return "+-" #both posible
        elif result[0]>r:
            return "+"
        elif result[0]<r:
            return "-"
        elif r==0:
            return "-"
        elif r==len(self.states)-1:
            return "+"

    def to_graph(self,thresh=0.01):
        sts =  self.find_states()
        g = graph()
        for i in range(self.no_genes):
            g.add_vert(i,"XYZUVWABCDEFGHIJKLMNOPQRST"[i])
        for i in range(self.no_genes):
            for j in range(self.no_genes):
                #print "krawedz",i,j
                rowne=True
                for s in self.st_spc.lst:
                    #print "stan",s
                    imgs=self.images(s,j)
                    for n in self.neighbours(s,i):
                        #print"stan2", n
                        imgn = self.images(n,j)
                        #print img
                        if imgn=="+-":
                            break
                        elif imgn!=imgs and imgs!="+-":
                            rowne=False
                            #print "jest krawedz",i,j,s,n,imgn
                            
                            break
                    if not rowne:
                        print i,"->",j
                        g.add_edge(i,j,imgn)
                        break
        return g

    def learn_observations(self,obs):
        return self.multiple_learn(obs,maxiter=2000)[1][-1]
